/*************************************************************************
caSCADA -- Data acquisition system program based on libmodbus 3.0.6
By Tony R. Kuphaldt
Copyright (C) 2016
Last update April 24, 2016

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

*************************************************************************/

#include <stdio.h>
#include <unistd.h>		// Necessary for the sleep() function
#include <modbus.h>		// Necessary for all modbus connection and data read/write functions
#include "cascada.h"		// Contains all the declarations specific to caSCADA

#define IP_ADDRESS "169.254.8.101"
#define TCP_PORT 502




int
main (void)
{

  // Define Modbus/TCP device and port
  Device = modbus_new_tcp (IP_ADDRESS, TCP_PORT);


  // What to do if this function returns an error
  if (Device == NULL)
    {
      fprintf (stderr, "Unable to establish TCP port \n");
      return -1;
    }


  // This very important function call instructs the libmodbus
  // library functions to attempt recovery in the event of
  // network interruptions.  Without it, the default behavior
  // of libmodbus is to do nothing at all, leaving any network
  // error recovery to the user!
  modbus_set_error_recovery(Device, MODBUS_ERROR_RECOVERY_LINK);

  // Establish a Modbus connection and what to do if this fails
  if (modbus_connect (Device) == -1)
    {
      fprintf (stderr, "Unable to connect to Modbus device \n");
      modbus_free (Device);
      return -1;
    }


// TO-DO!!!
// open cascada_poll_pointer.dat file ();



// Main loop
  while (1)
    {

      // Read all 14 analog input channels on the LabJack T7 DAQ.  This function happens to
      // store the 14 floating-point values in the global array "ain[]".
      read_LJT7_analogin ();


      // Call channel functions to assign values to the floating-point structure elements
      f_channel_00 ();
      f_channel_01 ();
      f_channel_02 ();
      f_channel_03 ();
      f_channel_04 ();
      f_channel_05 ();
      f_channel_06 ();
      f_channel_07 ();
      f_channel_08 ();
      f_channel_09 ();
      f_channel_10 ();
      f_channel_11 ();
      f_channel_12 ();
      f_channel_13 ();
      f_channel_14 ();
      f_channel_15 ();
      f_channel_16 ();
      f_channel_17 ();
      f_channel_18 ();
      f_channel_19 ();
      f_channel_20 ();
      f_channel_21 ();
      f_channel_22 ();
      f_channel_23 ();
      f_channel_24 ();
      f_channel_25 ();
      f_channel_26 ();
      f_channel_27 ();
      f_channel_28 ();
      f_channel_29 ();
      f_channel_30 ();
      f_channel_31 ();

      // Write structure data to text file ("data.txt")
      write_datatext ();

      // Write structure data to HTML file ("data.html")
      write_datahtml ();

      // Write 2 analog output channels on the LabJack T7 DAQ.  This function happens to
      // read from the 2 floating-point values in the global array "aout[]".
      write_LJT7_analogout ();

      // Wait for 1 second so we don't waste network bandwidth and processing power
      sleep (1);
    }



  // Close Modbus connection
  modbus_close (Device);


  // Free Modbus device and TCP port
  modbus_free (Device);


  // TO-DO!!!
  // close cascada_poll_pointer.dat file ();


  // Force of habit here -- I like all functions to return *something*
  return 1;

}









int
read_LJT7_analogin (void)
{

  uint16_t inreg_word[28];
  int read_count, n;

  union
  {
    uint16_t word[2];
    uint8_t byte[4];
  } in;

  union
  {
    uint8_t byte[4];
    float real;
  } out;


  // Read twenty-eight sequential 16-bit Modbus registers, starting with 0
  read_count = modbus_read_registers (Device, 0, 28, inreg_word);


  // If read command is unsuccessful . . .
  if (read_count < 1)
    {
      fprintf (stderr,
	       "No data read from DAQ -- network possibly severed! \n");
    }


  // Populate the ain[] 32-bit floating-point array from inreg_word[]
  // 16-bit integer data just read from the LabJack T7, swapping the
  // byte order as well
  for (n = 0; n < 14; ++n)
    {

      in.word[0] = inreg_word[2 * n];
      in.word[1] = inreg_word[(2 * n) + 1];
      /*
         The LabJack T7 stores its floating-point variables in a
         weird byte-order format that follows the pattern CDAB.
         This algorithm swaps the bytes as needed!
       */
      out.byte[0] = in.byte[2];
      out.byte[1] = in.byte[3];
      out.byte[2] = in.byte[0];
      out.byte[3] = in.byte[1];

      // Copy the union's data to the respective analog input array element
      ain[n] = out.real;
    }

  // Force of habit here -- I like all functions to return *something*
  return 1;
}





int
write_LJT7_analogout (void)
{

  uint16_t outreg_word[4];
  int write_count, n;

  union
  {
    uint8_t byte[4];
    float real;
  } in;

  union
  {
    uint16_t word[2];
    uint8_t byte[4];
  } out;


  // Populate the outreg_word[] 16-bit integer array with the 32-bit
  // floating-point values residing in the aout[] array, in preparation
  // for writing to the LabJack T7
  for (n = 0; n < 2; ++n)
    {

      // Safeguards against an unreasonably low value being written to
      // the analog output DAC, with 0.00 volts being the low limit
      if (aout[n] < 0)
	aout[n] = 0.00;

      // Safeguards against an unreasonably high value being written to
      // the analog output DAC, with 5.00 volts being the low limit
      if (aout[n] > 5.00)
	aout[n] = 5.00;

      // Copy the analog output array element to the union data
      in.real = aout[n];
      /*
         The LabJack T7 stores its floating-point variables in a
         weird byte-order format that follows the pattern CDAB.
         This algorithm swaps the bytes as needed!
       */
      out.byte[2] = in.byte[0];
      out.byte[3] = in.byte[1];
      out.byte[0] = in.byte[2];
      out.byte[1] = in.byte[3];

      outreg_word[2 * n] = out.word[0];
      outreg_word[(2 * n) + 1] = out.word[1];
    }


  // Write four sequential 16-bit Modbus registers, starting with 1000
  write_count = modbus_write_registers (Device, 1000, 4, outreg_word);


  // If write command is unsuccessful . . .
  if (write_count < 1)
    {
      fprintf (stderr,
	       "No data written to DAQ -- network possibly severed! \n");
    }


  // Force of habit here -- I like all functions to return *something*
  return 1;
}


