/*************************************************************************
caSCADA -- PID control system program based on libmodbus 3.0.6
By Tony R. Kuphaldt
Copyright (C) 2016
Last update April 24, 2016

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

This program is a self-contained application for PID control, using
analog input AIN0 as the process variable, and analog output DAC0 as
the manipulated variable.  An "ncurses" interface provides the user the
means to run and tune the controller.

*************************************************************************/

#include <stdio.h>
#include <math.h>		// Necessary for the fabs() function
#include <unistd.h>		// Necessary for the sleep() function
#include <modbus.h>		// Necessary for all modbus connection and data read/write functions
#include <ncurses.h>		// Necessary for the "ncurses" library functions for the operator display
#include <time.h>		// Necessary for the "time" library functions to work
#include "pid.h"		// Contains all the declarations specific to caSCADA's PID functionality

#define IP_ADDRESS "169.254.8.101"	// IPv4 address of the Modbus/TCP data acquisition unit
#define TCP_PORT 502		// TCP port number used for Modbus/TCP (502 is standard)

int
main (void)
{

  // Initialize all PID controller variables
  set_defaults ();


  // This function prints an introductory message to the
  // user's screen, prior to establishing a network
  // connection or executing the PID algorithm.  It returns
  // a value for "looprun" which (if not equal to 1) causes
  // the program to cleanly exit.
  looprun = splash_screen ();

  ///////////////////////////////////////////////////////
  //
  // These instructions set things up in preparation for
  // looping the PID algorithm.  We need to establish a
  // Modbus connection to the data acquisition unit, set
  // default control parameter values, and then open up
  // the "ncurses" display mode so we have a usable
  // operator interface.
  //
  ///////////////////////////////////////////////////////


  if (looprun == 1)
    {

      // Opens up Modbus network connection
      open_modbus_connection ();


      // Opens up ncurses mode and initializes all the important variables
      open_ncurses ();

    }

  // If looprun does not equal 1, exit the main routine NOW!
  else
    return 0;


  ///////////////////////////////////////////////////////
  //
  // This is the "main loop" of the program that gets
  // executed over and over again so long as "looprun"
  // is equal to a value of 1.  Otherwise, the program
  // cleanly exits.
  //
  ///////////////////////////////////////////////////////

  while (looprun)
    {

      // Increments the scan counter, counting the number of times this
      // loop has been executed
      ++scan_count;


      // Get the current UNIX system time
      time_current = (int) (time (NULL));


      // Read analog input AIN0 on the LabJack model T7 DAQ, 
      // using "CDAB" byte swapping (specified by the "2" argument)      
      ain[0] = read_32float(40001,2);  


      // Scales the process variable for PID loop 0 in percent,
      // from the 1-5 VDC input seen at analog input 0
      pid[0].PV = (ain[0] * 25) - 25;


      // Calculate the output of the PID algorithm
      pid_position (0);


      // Bounding trend interval to reasonable limits
      if (trend_interval < 1)
	trend_interval = 1;

      if (trend_interval > 9999)
	trend_interval = 9999;


      // Calls the trend-shifting function for PID loop 0
      // and plots trend data, every trend_interval number
      // of scans through the main while() loop
      if (scan_count % trend_interval == 0)
	trend_shift_plot (0);


      // Update the operator interface display for PID loop 0
      display_plot (0);


      // Read keyboard key strokes from the human operator
      // for PID loop 0
      keyboard_scan (0);


      // Scales the output variable for PID loop 0 into a
      // DC voltage range on the LabJack's analog output
      aout[0] = (pid[0].OUT / 25) + 1;	// Scales to 1-5 VDC
      // aout[0] = pid[0].OUT / 20;     // Scales to 0-5 VDC

      // Safeguards against an unreasonably low value being written to
      // the analog output DAC, with 0.00 volts being the low limit
      if (aout[0] < 0)
	aout[0] = 0.00;

      // Safeguards against an unreasonably high value being written to
      // the analog output DAC, with 5.00 volts being the low limit
      if (aout[0] > 5.00)
	aout[0] = 5.00;

      // Write analog input DAC0 on the LabJack model T7 DAQ, 
      // using "CDAB" byte swapping (specified by the "2" argument)      
      write_32float(aout[0], 41001, 2);

      // Update the last scan time value
      time_lastscan = time_current;

    }


  ///////////////////////////////////////////////////////////
  //
  //  These instructions are executed only when the while()
  //  loop stops, which is triggered by looprun = 0.  Their
  //  purpose is to cleanly exit the program, including all
  //  the "housekeeping" instructions necessary to get out
  //  of ncurses mode and return to normal terminal mode,
  //  as well as closing any Modbus network connections.
  //
  ///////////////////////////////////////////////////////////

  // Closes out ncurses mode and returns to normal terminal mode
  close_ncurses ();


  // Closes out Modbus network connection
  close_modbus_connection ();


  // Force of habit here -- I like all functions to return *something*
  return 1;

}







int
open_modbus_connection (void)
{

  // Define Modbus/TCP device and port
  Device = modbus_new_tcp (IP_ADDRESS, TCP_PORT);


  // What to do if this function returns an error
  if (Device == NULL)
    {
      fprintf (stderr, "Unable to establish TCP port \n");
      looprun = 0;		// This will force the program to exit
      return -1;
    }


  // This very important function call instructs the libmodbus
  // library functions to attempt recovery in the event of
  // network interruptions.  Without it, the default behavior
  // of libmodbus is to do nothing at all, leaving any network
  // error recovery to the user!
  modbus_set_error_recovery (Device, MODBUS_ERROR_RECOVERY_LINK);


  while (modbus_connect (Device) == -1)
    {
      fprintf (stderr, "Trying to connect to the Modbus device\n");
      // sleep (2);
    }


  // Force of habit here -- I like all functions to return *something*
  return 1;
}






int
close_modbus_connection (void)
{
  // Close Modbus connection
  modbus_close (Device);

  // Free Modbus device and TCP port
  modbus_free (Device);

  // Force of habit here -- I like all functions to return *something*
  return 1;
}







int
open_ncurses (void)
{

  // Starts ncurses mode (the console-based operator interface display)
  initscr ();

  // This ncurses function disables line buffering, which makes pressing "Enter" unnecessary
  // raw();   // raw() disables all normal keyboard control sequences
  cbreak ();			// cbreak() allows control sequences such as Ctrl-C to still work like normal

  // This ncurses function gives us access to Function keys (F1, F2, etc.) on the keyboard
  keypad (stdscr, TRUE);

  // Don't echo characters to the screen whenever the getch() function is called to accept keyboard input
  noecho ();

  // Don't halt the program execution when the getch() function is called
  nodelay (stdscr, TRUE);

  curs_set (0);			// Makes the cursor invisible

  // Test to see if our ncurses environment supports colors, which is very nice
  if (has_colors () == FALSE)
    {
      endwin ();
      printf ("Sorry, but colors aren't supported in your terminal! \n");
      return 0;
    }

  // Starts ncurses color capability
  start_color ();

  // Defining text/background color pair combinations for ncurses display
  //
  // Color pair "0" is default: WHITE text on a BLACK background
  //
  // Defines color pair "1" as BLACK text on a GREEN background
  init_pair (1, COLOR_BLACK, COLOR_GREEN);
  //
  // Defines color pair "11" as WHITE text on a GREEN background
  init_pair (11, COLOR_WHITE, COLOR_GREEN);
  //
  // Defines color pair "2" as BLACK text on a YELLOW background
  init_pair (2, COLOR_BLACK, COLOR_YELLOW);
  //
  // Defines color pair "22" as WHITE text on a YELLOW background
  init_pair (22, COLOR_WHITE, COLOR_YELLOW);
  //
  // Defines color pair "3" as BLACK text on a WHITE background
  init_pair (3, COLOR_BLACK, COLOR_WHITE);
  //
  // Defines color pair "4" as RED text on a BLACK background
  init_pair (4, COLOR_RED, COLOR_BLACK);
  //
  // Defines color pair "44" as RED text on a WHITE background
  init_pair (44, COLOR_RED, COLOR_WHITE);


  // Force of habit here -- I like all functions to return *something*
  return 1;
}





int
close_ncurses (void)
{
  endwin ();

  // Force of habit here -- I like all functions to return *something*
  return 1;
}











int
pid_position (int c)
{

  // Declaring some of the variables used within this function

  float error;			// Difference between PV and SP
  float derivative;		// Rate-of-change with respect to time
  int integral_inhibit;		// 0 = Allow integration ; 1 = Stop integration
  static int scan_count_last;	// Scan count from the last scan
  static float PV_history[10];	// Array storing the last several PV values
  int n;


  // Bounding process variable between limits of -5% and +105%
  if (pid[c].PV > 105)
    pid[c].PV = 105;

  if (pid[c].PV < -5)
    pid[c].PV = -5;

  PV_history[0] = pid[c].PV;

  // Bounding setpoint between limits of -5% and +105%
  if (pid[c].SP > 105)
    pid[c].SP = 105;

  if (pid[c].SP < -5)
    pid[c].SP = -5;


  // Bounding tuning coefficients between reasonable limits
  if (pid[c].K_P > 50)
    pid[c].K_P = 50;

  if (pid[c].K_P < 0)
    pid[c].K_P = 0;

  if (pid[c].K_I > 99)
    pid[c].K_I = 99;

  if (pid[c].K_I < 0)
    pid[c].K_I = 0;

  if (pid[c].K_D > 99)
    pid[c].K_D = 99;

  if (pid[c].K_D < 0)
    pid[c].K_D = 0;

  if (pid[c].I_db > 9.9)
    pid[c].I_db = 9.9;

  if (pid[c].I_db < 0)
    pid[c].I_db = 0;



  // Calculating program loop scan rate.  This is important for accurate
  // calculation of time-based control elements such as integral and
  // derivative (the "I" and "D" terms of the PID algorithm).  Since the
  // operating system is *not* executing PID on a locked-time schedule
  // as is the case with most industrial control systems, we need our
  // program to determine how long each "loop" takes and use that live
  // scan-time value to compensate for anything that slows the processor
  // down (e.g. file read/write operations, user logins, etc.).

  if (time_current != time_lastscan)	// Every time the system clock increments
    {
      scans_per_second =
	(float) (scan_count - scan_count_last) / (float) (time_current -
							  time_lastscan);
      scan_count_last = scan_count;

      // This bounds scans_per_second to a reasonable value in case we get
      // a weird number (e.g. when the program first starts up)
      if (scans_per_second < 0.1)
	scans_per_second = 0.1;
    }


  // Automatic mode
  if (pid[c].am_mode == 1)
    {

      // Calculating error, reverse action
      if (pid[c].action == 0)
	error = pid[c].SP - pid[c].PV;

      // Calculating error, direct action
      else
	error = pid[c].PV - pid[c].SP;


      // The following if/else conditionals will inhibit integral (reset) action
      // under certain conditions, such as the output exceeding windup limits, or
      // exceeding saturation limits (0 and 100%), or if the error between PV and
      // SP is less than the "integral deadband" value.  The latter condition is
      // useful in control applications where the final control element
      // (e.g. control valve, or motor-operated device) is incapable of precise
      // positioning, which will cause integral action to ceaselessly ramp up and
      // down in a futile effort to precisely achieve setpoint.  Integral deadband
      // simply says that getting close to setpoint is good enough, and stops the
      // integral action within that distance of PV.

      // Set inhibit = 1 if we try to exceed the high windup limit or 100 percent
      if (((pid[c].OUT > pid[c].windup_hilimit) || (pid[c].OUT > 100))
	  && error > 0)
	integral_inhibit = 1;
      // Set inhibit = 1 if we try to go below the low windup limit or 0 percent
      else if (((pid[c].OUT < pid[c].windup_lolimit) || (pid[c].OUT < 0))
	       && error < 0)
	integral_inhibit = 1;
      // Set inhibit = 1 if error is less than the integral deadband
      else if (fabs (error) < fabs (pid[c].I_db))
	integral_inhibit = 1;
      else
	integral_inhibit = 0;	// Otherwise, let integration occur


      // Integral summation, accumulated as the BIAS value as long as it is not inhibited.
      // For each scan, the BIAS value gets incremented or decremented
      // by an amount equal to the error times the difference in scan times (delta t)
      // times the integral coefficient divided by 60 (converting rpm to rps).
      // For the ideal algorithm we also multiply by K_P.

      if (integral_inhibit == 0)
	{
	  if (pid[c].algorithm == 0)	// Ideal algorithm
	    pid[c].BIAS =
	      pid[c].BIAS +
	      (pid[c].K_P * error * pid[c].K_I / (60 * scans_per_second));

	  if (pid[c].algorithm == 1)	// Parallel algorithm
	    pid[c].BIAS =
	      pid[c].BIAS + (error * pid[c].K_I / (60 * scans_per_second));
	}


      // Derivative calculation, based on rate-of-change of PV.
      // In order to avoid differentiating over unreasonably short
      // time intervals (where unavoidable jumps in the PV stemming
      // from noise or digital count value changes result in huge
      // dPV/dt values), we will calculate the rate of PV change
      // over multiple scans, using the PV_history[] array which
      // is refreshed with every scan of this PID function.
      // PV_history[0] is the current PV value, while
      // PV_history[9] is the oldest.  We can set the dt interval
      // as wide as we like (PV_history[9] - PV_history[0]) or as
      // narrow as we like (PV_history[1] - PV_history[0]).  The
      // wider the interval (i.e. greater dt), the greater the
      // divider constant must be following "scans_per_second"

      if (pid[c].algorithm == 0)	// Ideal algorithm assuming reverse action
	derivative =
	  pid[c].K_P * pid[c].K_D * (PV_history[2] -
				     PV_history[0]) * scans_per_second / 2;

      if (pid[c].algorithm == 1)	// Parallel algorithm assuming reverse action
	derivative =
	  pid[c].K_D * (PV_history[2] - PV_history[0]) * scans_per_second / 2;

      if (pid[c].action == 1)	// Swaps sign of derivative term if direct action
	derivative = -derivative;


      ////////////////////////////////////////////////////////////////
      // This section of the function puts the P, I, and D terms together
      ////////////////////////////////////////////////////////////////

      // P+I+D algorithm
      pid[c].OUT = (pid[c].K_P * error) + pid[c].BIAS + derivative;

    }


  // Manual mode
  if (pid[c].am_mode == 0)
    {
      pid[c].SP = pid[c].PV;	// This provides setpoint tracking in manual mode
      pid[c].OUT = pid[c].BIAS;
    }


  // Bounding output between limits of -5% and +105%
  if (pid[c].OUT > 105)
    pid[c].OUT = 105;

  if (pid[c].OUT < -5)
    pid[c].OUT = -5;

  // Updates "PV_history" array with every scan of this function
  for (n = 9; n > 0; --n)
    {
      PV_history[n] = PV_history[n - 1];
    }

  // Force of habit here -- I like all functions to return *something*
  return 1;
}






int
display_plot (int c)
{

  int blink_color;

  ////////////////////////////////////////////////////////////////
  //
  // This function prints static and numerical data to screen
  //
  ////////////////////////////////////////////////////////////////

  // Using the attron() and attroff() functions to control ncurses coloring
  // Using the move() and printw() functions to place text on the screen
  //   syntax: move( y , x ) where 0,0 is the upper-left corner

  // Placing the title and function key legend at the program start,
  // rather than re-placing all this static text with every single
  // call of the display_plot() function (this saves times).
  if (scan_count < 2)
    {
      attron (COLOR_PAIR (3));
      move (3, 29);
      printw ("caSCADA PID version 2.3");
      move (4, 29);
      printw ("   Copyright (C) 2016  ");
      move (5, 29);
      printw ("    GNU GPL version 3  ");
      move ((TRENDHEIGHT + 11), 1);
      printw ("F1 (M) Manual");
      move ((TRENDHEIGHT + 12), 1);
      printw ("F2 (A) Auto  ");
      move ((TRENDHEIGHT + 14), 48);
      printw ("F4 (pgdwn) - 10 ");
      move ((TRENDHEIGHT + 13), 48);
      printw ("F5 (down)  - 1  ");
      move ((TRENDHEIGHT + 12), 48);
      printw ("F6 (left)  - 0.1");
      move ((TRENDHEIGHT + 11), 48);
      printw ("F7 (right) + 0.1");
      move ((TRENDHEIGHT + 10), 48);
      printw ("F8  (up)   + 1  ");
      move ((TRENDHEIGHT + 9), 48);
      printw ("F9 (pgup)  + 10 ");
      move ((TRENDHEIGHT + 9), 67);
      printw ("F11 (S) Select");	// This function key changes the select_mode variable
      move ((TRENDHEIGHT + 11), 71);
      printw ("(Q) Exit");
      move ((TRENDHEIGHT + 13), 66);
      printw ("F12 (T) Capture");
      move ((TRENDHEIGHT + 14), 66);
      printw ("  trend data   ");
      attroff (COLOR_PAIR (3));
    }


  // Placing the PV, SP, and OUT numerical displays
  attron (COLOR_PAIR (1));
  move (0, 1);
  printw ("PV = %3.1f percent = %f %s  ", pid[c].PV,
	  (pid[c].PV / 100) * (pid[c].URV - pid[c].LRV) + pid[c].LRV,
	  pid[c].UNIT);

  if (select_mode == 0 && pid[c].am_mode == 1)
    attron (COLOR_PAIR (11));
  else
    attron (COLOR_PAIR (1));
  move (1, 1);
  printw ("SP = %3.1f percent = %f %s  ", pid[c].SP,
	  (pid[c].SP / 100) * (pid[c].URV - pid[c].LRV) + pid[c].LRV,
	  pid[c].UNIT);

  if (select_mode == 0 && pid[c].am_mode == 0)
    attron (COLOR_PAIR (11));
  else
    attron (COLOR_PAIR (1));
  move (2, 1);
  printw ("OUT = %3.1f percent ", pid[c].OUT);
  attroff (COLOR_PAIR (1));




  // Placing the PID scan execution rate display
  attron (COLOR_PAIR (1));
  move (5, 1);
  printw ("PID scan rate = %.1f/sec ", scans_per_second);
  attroff (COLOR_PAIR (1));



  // PV alarm displays, placement and coloring

  // For high-high and low-low alarm conditions, we want the alarm
  // display to alternate background colors (white/black) to better
  // grab the user's attention.  This pair of if/else statements
  // makes the "blink_color" variable switch between values of 4
  // and 44 which are the color pair codes for red on black versus
  // red on white.
  if (time_current % 2 == 0)
    blink_color = 44;

  else
    blink_color = 4;



  if (pid[c].PV > pid[c].PV_hihi)
    {
      attron (COLOR_PAIR (blink_color));
      move (4, 1);
      printw ("PV high-high alarm!");
      attroff (COLOR_PAIR (blink_color));
    }

  else if (pid[c].PV > pid[c].PV_hi)
    {
      attron (COLOR_PAIR (44));
      move (4, 1);
      printw ("PV high alarm!     ");
      attroff (COLOR_PAIR (44));
    }

  else if (pid[c].PV < pid[c].PV_lolo)
    {
      attron (COLOR_PAIR (blink_color));
      move (4, 1);
      printw ("PV low-low alarm!  ");
      attroff (COLOR_PAIR (blink_color));
    }

  else if (pid[c].PV < pid[c].PV_lo)
    {
      attron (COLOR_PAIR (44));
      move (4, 1);
      printw ("PV low alarm!      ");
      attroff (COLOR_PAIR (44));
    }

  else
    {
      attron (COLOR_PAIR (0));
      move (4, 1);
      printw ("                   ");
      attroff (COLOR_PAIR (0));
    }




  // Placing the modes and tuning parameters display, the
  // color scheme of each parameter switching based on the
  // value of select_mode.  This shows the user which
  // parameter they are set to adjust with the increment/
  // decrement controls.
  if (select_mode == 0)
    attron (COLOR_PAIR (22));
  else
    attron (COLOR_PAIR (2));
  move (0, 54);
  if (pid[c].am_mode == 0)
    printw ("Man -- (adjust output)    ");
  else
    printw ("Auto -- (adjust setpoint) ");

  if (select_mode == 1)
    attron (COLOR_PAIR (22));
  else
    attron (COLOR_PAIR (2));
  move (1, 54);
  printw ("K_P = %1.1f (gain)          ", pid[c].K_P);
  attroff (COLOR_PAIR (2));

  if (select_mode == 2)
    attron (COLOR_PAIR (22));
  else
    attron (COLOR_PAIR (2));
  move (2, 54);
  printw ("K_I = %1.1f r/min           ", pid[c].K_I);
  attroff (COLOR_PAIR (2));

  if (select_mode == 3)
    attron (COLOR_PAIR (22));
  else
    attron (COLOR_PAIR (2));
  move (2, 71);
  printw ("DB = %1.1f%%", pid[c].I_db);
  attroff (COLOR_PAIR (2));

  if (select_mode == 4)
    attron (COLOR_PAIR (22));
  else
    attron (COLOR_PAIR (2));
  move (3, 54);
  printw ("K_D = %1.1f seconds         ", pid[c].K_D);
  attroff (COLOR_PAIR (2));

  if (select_mode == 5)
    attron (COLOR_PAIR (22));
  else
    attron (COLOR_PAIR (2));
  move (4, 54);
  if (pid[c].action == 0)
    printw ("Reverse-acting,   ");
  else
    printw ("Direct-acting,    ");
  attroff (COLOR_PAIR (2));

  if (select_mode == 6)
    attron (COLOR_PAIR (22));
  else
    attron (COLOR_PAIR (2));
  move (4, 71);
  if (pid[c].algorithm == 0)
    printw ("Ideal    ");
  else
    printw ("Parallel ");
  attroff (COLOR_PAIR (2));

  if (select_mode == 7)
    attron (COLOR_PAIR (22));
  else
    attron (COLOR_PAIR (2));
  move (5, 54);
  printw ("Trend interval = %3i scans", trend_interval);
  attroff (COLOR_PAIR (2));


  // Calculating the timebase of the trend graph in real time units
  // (number of ticks per second of real time)
  timebase = scans_per_second / trend_interval;

  // Placing the trend timebase display
  attron (COLOR_PAIR (1));
  move ((TRENDHEIGHT + 9), 1);
  if (timebase >= 1)
    printw ("Trend update rate = %.1f ticks per second  ", timebase);
  else
    printw ("Trend update rate = %.1f seconds per tick  ", 1 / timebase);
  attroff (COLOR_PAIR (1));


  // The refresh() ncurses function prints all the preceding text to the screen
  refresh ();


  // Force of habit here -- I like all functions to return *something*
  return 1;
}









int
keyboard_scan (int c)
{

  int key;
  float step = 0.0;

  ///////////////////////////////////////////////////////
  // This function reads keyboard input with each call,
  // without halting the execution of the PID loop.
  ///////////////////////////////////////////////////////

  key = getch ();

  switch (key)
    {
      // F1 or M switches to manual mode
    case KEY_F (1):
    case 'M':
    case 'm':
      pid[c].am_mode = 0;
      pid[c].BIAS = pid[c].OUT;
      select_mode = 0;		// Switch to mode 0 (OUT adjust) when changing to manual mode
      break;

      // F2 or A switches to Automatic mode
    case KEY_F (2):
    case 'A':
    case 'a':
      pid[c].am_mode = 1;
      select_mode = 0;		// Switch to mode 0 (SP adjust) when changing to automatic mode
      break;

      // F4 or Page Down (Next Page) decrements by -10
    case KEY_F (4):
    case KEY_NPAGE:
      step = -10;
      break;

      // F5 or down arrow decrements by -1
    case KEY_F (5):
    case KEY_DOWN:
      step = -1;
      break;

      // F6 or left arrow decrements by -0.1
    case KEY_F (6):
    case KEY_LEFT:
      step = -0.1;
      break;

      // F7 or right arrow increments by +0.1
    case KEY_F (7):
    case KEY_RIGHT:
      step = 0.1;
      break;

      // F8 or up arrow increments by +1
    case KEY_F (8):
    case KEY_UP:
      step = 1;
      break;

      // F9 or Page Up (Previous Page) increments by +10
    case KEY_F (9):
    case KEY_PPAGE:
      step = 10;
      break;

      // F11 or S selects parameter to change
    case KEY_F (11):
    case 'S':
    case 's':
      select_mode = tuning_entry ();
      break;

      // F12 or T captures the trend graph data and saves to a comma-delimited text file (.csv)
    case KEY_F (12):
    case 'T':
    case 't':
      capture_trend ();
      break;

      // Q cleanly exits the program
    case 'Q':
    case 'q':
      looprun = 0;
      break;
    }



  // These conditional statements declare what to do with the increment/decrement
  // variable "step" depending on the mode of the controller
  if (pid[c].am_mode == 0 && select_mode == 0)
    {
      pid[c].BIAS = pid[c].BIAS + step;	// When in manual mode, change the BIAS value

      // Bounding BIAS between limits of -5% and +105% when in manual mode
      if (pid[c].BIAS > 105)
	pid[c].BIAS = 105;
      if (pid[c].BIAS < -5)
	pid[c].BIAS = -5;
    }

  if (pid[c].am_mode == 1 && select_mode == 0)
    pid[c].SP = pid[c].SP + step;	// When in automatic mode, change the SP value

  if (select_mode == 1)
    pid[c].K_P = pid[c].K_P + step;	// Adjusts proportional gain coefficient

  if (select_mode == 2)
    pid[c].K_I = pid[c].K_I + step;	// Adjusts integral coefficient

  if (select_mode == 3)
    pid[c].I_db = pid[c].I_db + step;	// Adjusts integral deadband

  if (select_mode == 4)
    pid[c].K_D = pid[c].K_D + step;	// Adjusts derivative coefficient

  if (select_mode == 5 && pid[c].am_mode == 0)	// Only allows change of action when in manual mode!
    {
      if (step > 0)		// Sets direct action
	pid[c].action = 1;

      if (step < 0)		// Sets reverse action
	pid[c].action = 0;
    }

  if (select_mode == 6 && pid[c].am_mode == 0)	// Only allows change of algorithm when in manual mode!
    {
      if (step > 0)		// Sets "parallel" algorithm
	pid[c].algorithm = 1;

      if (step < 0)		// Sets "ideal" algorithm
	pid[c].algorithm = 0;
    }

  if (select_mode == 7)
    trend_interval = trend_interval + (int) (step);	// Adjusts trend scan interval


  // Force of habit here -- I like all functions to return *something*
  return 1;
}







int
set_defaults (void)
{

  // PID controller 0
  pid[0].SP = 50;		// Initialize setpoint to 50%
  pid[0].BIAS = 50;		// Initialize bias to 50%
  pid[0].K_P = 1;		// Initialize gain at a value of 1
  pid[0].K_I = 0;		// Zero repeats per minute integral
  pid[0].K_D = 0;		// Zero seconds derivative
  pid[0].PV_hihi = 95;
  pid[0].PV_hi = 90;
  pid[0].PV_lo = 10;
  pid[0].PV_lolo = 5;
  pid[0].URV = 2200;
  pid[0].LRV = 0;
  pid[0].UNIT = "RPM";
  pid[0].windup_hilimit = 99;	// Integral windup high limit
  pid[0].windup_lolimit = 1;	// Integral windup low limit
  pid[0].I_db = 0;		// Integral deadband (integration halts at error values less than this)
  pid[0].action = 0;		// 0 = Reverse      ;  1 = Direct
  pid[0].am_mode = 0;		// 0 = Manual       ;  1 = Automatic
  pid[0].algorithm = 0;		// 0 = Ideal        ;  1 = Parallel

  /********************************************************
   *
   * The following code in this function is necessary only
   * when multiple PID loops are being executed, such as in
   * the case of a cascade control system.
   *

  // PID controller 1
  pid[1].SP = 50;		// Initialize setpoint to 50%
  pid[1].BIAS = 50;		// Initialize bias to 50%
  pid[1].K_P = 1;               // Initialize gain at a value of 1
  pid[1].K_I = 0;               // Zero repeats per minute integral
  pid[1].K_D = 0;               // Zero seconds derivative
  pid[1].PV_hihi = 100;
  pid[1].PV_hi = 95;
  pid[1].PV_lo = 5;
  pid[1].PV_lolo = 0;
  pid[1].URV = 100;
  pid[1].LRV = 0;
  pid[1].UNIT = "percent";
  pid[1].windup_hilimit = 99;   // Integral windup high limit
  pid[1].windup_lolimit = 1;    // Integral windup low limit
  pid[1].I_db = 0;              // Integral deadband (integration halts at error values less than this)
  pid[1].action = 0;		// 0 = Reverse      ;  1 = Direct
  pid[1].am_mode = 0;		// 0 = Manual       ;  1 = Automatic
  pid[1].algorithm = 0;		// 0 = Ideal        ;  1 = Parallel

  // PID controller 2
  pid[2].SP = 50;		// Initialize setpoint to 50%
  pid[2].BIAS = 50;		// Initialize bias to 50%
  pid[2].K_P = 1;               // Initialize gain at a value of 1
  pid[2].K_I = 0;               // Zero repeats per minute integral
  pid[2].K_D = 0;               // Zero seconds derivative
  pid[2].PV_hihi = 100;
  pid[2].PV_hi = 95;
  pid[2].PV_lo = 5;
  pid[2].PV_lolo = 0;
  pid[2].URV = 100;
  pid[2].LRV = 0;
  pid[2].UNIT = "percent";
  pid[2].windup_hilimit = 99;   // Integral windup high limit
  pid[2].windup_lolimit = 1;    // Integral windup low limit
  pid[2].I_db = 0;              // Integral deadband (integration halts at error values less than this)
  pid[2].action = 0;		// 0 = Reverse      ;  1 = Direct
  pid[2].am_mode = 0;		// 0 = Manual       ;  1 = Automatic
  pid[2].algorithm = 0;		// 0 = Ideal        ;  1 = Parallel

  ********************************************************/

  trend_interval = 5;

  // Force of habit here -- I like all functions to return *something*
  return 1;
}








int
tuning_entry (void)
{

  // select_mode values:
  // 0 = operating mode -- adjusting OUT in manual or SP in automatic
  // 1 = adjust K_P
  // 2 = adjust K_I
  // 3 = adjust integral deadband
  // 4 = adjust K_D
  // 5 = adjust action (direct/reverse)
  // 6 = adjust algorithm (ideal/parallel)
  // 7 = adjust trend interval

  if (select_mode == 7)
    return 0;

  else if (select_mode == 6)
    return 7;

  else if (select_mode == 5)
    return 6;

  else if (select_mode == 4)
    return 5;

  else if (select_mode == 3)
    return 4;

  else if (select_mode == 2)
    return 3;

  else if (select_mode == 1)
    return 2;

  else if (select_mode == 0)
    return 1;

  else
    return 0;

}









int
trend_shift_plot (int c)
{

  int m, n;

  // This for() loop shifts data in all three trend arrays, to make the
  // trend points "march" across the screen

  // The 0th element of each array is the most recent, while the TRENDWIDTH-1
  // element is the oldest.  The following "for" loop takes data from the n-1
  // element and copies that data to the n element, thus updating all the
  // older elements with information from the newer.

  for (n = (TRENDWIDTH - 1); n > 0; --n)
    {
      pen[n].PV = pen[n - 1].PV;
      pen[n].SP = pen[n - 1].SP;
      pen[n].OUT = pen[n - 1].OUT;
    }

  // After shifting all the old data, we populate the 0th element with
  // fresh data from the pid[c] structure:

  pen[0].PV = pid[c].PV;
  pen[0].SP = pid[c].SP;
  pen[0].OUT = pid[c].OUT;

  // Bounds trend data between 0 and 100 percent so that nothing can be
  // drawn outside the borders of the trend graph
  if (pen[0].PV < 0)
    pen[0].PV = 0;

  if (pen[0].PV > 100)
    pen[0].PV = 100;

  if (pen[0].SP < 0)
    pen[0].SP = 0;

  if (pen[0].SP > 100)
    pen[0].SP = 100;

  if (pen[0].OUT < 0)
    pen[0].OUT = 0;

  if (pen[0].OUT > 100)
    pen[0].OUT = 100;


  /////////////////////////////////////////////////////////////////////
  //
  // This section of the function prints graphical trend data to screen
  //
  /////////////////////////////////////////////////////////////////////

  // Trend graph border:
  //   upper-left corner coordinates (y,x)  = 7 , 1
  //   lower-right corner coordinates (y,x) = TRENDHEIGHT+2 +7 , TRENDWIDTH+2
  // This leaves a trend graph vertical span from y=8 to y=TRENDHEIGHT+8
  // and a horizontal span from x=2 to x=TRENDWIDTH+1

  // Draws upper border of trend graph
  attron (COLOR_PAIR (0));
  for (n = 1; n < (TRENDWIDTH + 3); ++n)
    {
      move (7, n);
      printw ("-");
    }

  // Draws lower border of trend graph
  for (n = 1; n < (TRENDWIDTH + 3); ++n)
    {
      move (TRENDHEIGHT + 8, n);
      printw ("-");
    }

  // Draws left border of trend graph
  for (n = 7; n < (TRENDHEIGHT + 8); ++n)
    {
      move (n, 1);
      printw ("|");
    }

  // Draws right border of trend graph
  for (n = 7; n < (TRENDHEIGHT + 8); ++n)
    {
      move (n, (TRENDWIDTH + 2));
      printw ("|");
    }

  // Draws border corners of trend graph
  move (7, 1);
  printw ("+");
  move (7, (TRENDWIDTH + 2));
  printw ("+");
  move ((TRENDHEIGHT + 8), 1);
  printw ("+");
  move ((TRENDHEIGHT + 8), (TRENDWIDTH + 2));
  printw ("+");

  // Draws blank trend screen background.  This solution,
  // crude as it is, actually occupies fewer processor
  // cycles and makes for a more readable display than
  // using the ncurses clear() function!
  for (n = 8; n < (TRENDHEIGHT + 8); ++n)
    {
      for (m = 2; m < (TRENDWIDTH + 2); ++m)
	{
	  move (n, m);
	  printw (" ");		// Prints an empty space (BLACK) to the screen
	}
    }

  attroff (COLOR_PAIR (0));


  // Now we plot pen data to the screen

  attron (COLOR_PAIR (0));

  for (n = 0; n < TRENDWIDTH; ++n)
    {
      move ((TRENDHEIGHT + 7) -
	    (int) (((pen[n].OUT + (50 / TRENDHEIGHT)) / 100) * (TRENDHEIGHT -
								1)),
	    (TRENDWIDTH + 1) - n);
      printw ("o");

      move ((TRENDHEIGHT + 7) -
	    (int) (((pen[n].SP + (50 / TRENDHEIGHT)) / 100) * (TRENDHEIGHT -
							       1)),
	    (TRENDWIDTH + 1) - n);
      printw ("s");

      move ((TRENDHEIGHT + 7) -
	    (int) (((pen[n].PV + (50 / TRENDHEIGHT)) / 100) * (TRENDHEIGHT -
							       1)),
	    (TRENDWIDTH + 1) - n);
      printw ("p");
    }

  attroff (COLOR_PAIR (0));


  // The refresh() ncurses function prints all the preceding text to the screen
  refresh ();


  // Force of habit here -- I like all functions to return *something*
  return 1;
}








int
capture_trend (void)
{

  int n;
  FILE *fp;			// Pointer to a FILE type


  // opening the data file "0_pid_trend.csv" write-only
  if ((fp = fopen ("0_pid_trend.csv", "w")) == NULL)
    {
      fprintf (stderr, "Can't open `0_pid_trend.csv' text file \n");	// Do this if the file doesn't exist
      return 2;			// Return value of "2" means we cannot open the file
    }


  // Prints a "header" line to label each column of data
  fprintf (fp, "Sample , PV , SP , OUT \n");


  // Write pen data to trend capture file
  for (n = (TRENDWIDTH - 1); n > 0; --n)
    {
      fprintf (fp, "%i , %f , %f , %f \n", TRENDWIDTH - n, pen[n].PV,
	       pen[n].SP, pen[n].OUT);
    }


  // Close the "0_pid_trend.csv" text file
  fclose (fp);


  // Force of habit here -- I like all functions to return *something*
  return 1;
}








int
splash_screen (void)
{
  char key;
  int m, n;

  for (n = 0; n < 50; ++n)
    {
      printf ("\n");
    }

  printf ("\n");
  printf
    ("Welcome to the caSCADA PID controller!  This program implements \n");
  printf
    ("Proportional-Integral-Derivative (PID) control with a simple user \n");
  printf ("interface. \n");

  printf ("\n");
  printf ("Copyright (C) 2016 Tony R. Kuphaldt \n");
  printf ("This program comes with ABSOLUTELY NO WARRANTY! \n");
  printf ("This is free software, and you are welcome to distribute it \n");
  printf
    ("under certain conditions; view the file `GPL_v3.txt' for details. \n");

  printf ("\n");
  printf
    ("Edit the file `pid.c' to change these network parameters, found as #define statements: \n\n");
  printf ("   LabJack IP address = %s \n", IP_ADDRESS);
  printf ("   TCP port number = %i \n", TCP_PORT);

  printf ("\n");
  printf
    ("Edit the file `pid.c' to change these range parameters, found in the set_defaults() function: \n\n");
  printf
    ("   Process variable lower range value = %f   High-high alarm = %f   High alarm = %f \n",
     pid[0].LRV, pid[0].PV_hihi, pid[0].PV_hi);
  printf
    ("   Process variable upper range value = %f   Low-Low alarm = %f   Low alarm = %f \n",
     pid[0].URV, pid[0].PV_lolo, pid[0].PV_lo);
  printf ("   Process variable engineering units = %s \n", pid[0].UNIT);

  printf ("\n");
  printf ("This is the size of the trend display (%i wide by %i high):\n",
	  TRENDWIDTH, TRENDHEIGHT);
  printf ("(edit the file pid.h to change)\n\n");

  printf ("+");

  for (m = 0; m < TRENDWIDTH; ++m)
    printf ("-");

  printf ("+\n");

  for (n = 0; n < TRENDHEIGHT; ++n)
    {
      printf ("|");

      for (m = 0; m < TRENDWIDTH; ++m)
	printf (" ");

      printf ("|\n");
    }

  printf ("+");

  for (m = 0; m < TRENDWIDTH; ++m)
    printf ("-");

  printf ("+\n");

  printf ("\n");
  printf
    ("If any of these default settings is incorrect, please quit now and\n");
  printf ("edit the source code to your liking.\n");

  printf ("\nEnter Y to continue, anything else to quit ");
  scanf ("%c", &key);

  if (key == 'Y' || key == 'y')
    return 1;

  else
    return 0;

}
