/*************************************************************************
caSCADA -- Data acquisition system program based on libmodbus 3.0.6
By Tony R. Kuphaldt
Copyright (C) 2016
Last update April 24, 2016

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


*************************************************************************/



/**************************************************************************** 
		Declare Modbus device pointer 
****************************************************************************/

// Declaring Modbus/TCP device pointer
modbus_t *Device;




/**************************************************************************** 
		Declare PID structure
****************************************************************************/

// "data_pid" is the type of structure.  Here we define 3 of them, each one
// named "pid" followed by a number in brackets distinguishing the 3
// structures from each other.  That number may be thought of being analogous
// to subscripts for mathematical variables (e.g. x_1, x_2).
// This data structure holds all the variables that each PID algorithm needs
// in order to properly function.

struct data_pid
{
  float PV;                     // Process Variable
  float SP;                     // Setpoint
  float OUT;                    // Output
  float BIAS;                   // Bias (also used as the integral term)
  float K_P;                    // Proportional (gain) tuning coefficient
  float K_I;                    // Integral (reset) tuning coefficient
  float K_D;                    // Derivative (rate) tuning coefficient
  float PV_hihi;                // Hi-Hi alarm threshold 
  float PV_hi;                  // Hi alarm threshold
  float PV_lo;                  // Lo alarm threshold
  float PV_lolo;                // Lo-Lo alarm threshold
  float URV;                    // Upper Range Value for PV and SP (engineering units)
  float LRV;                    // Lower Range Value for PV and SP (engineering units) 
  char *UNIT;                   // Units of measurement for LRV and URV
  float windup_hilimit;         // Hi limit where integral action is told to halt
  float windup_lolimit;         // Lo limit where integral action is told to halt
  float I_db;                   // Integral deadband (integration halts at error values less than this)
  int action;			// 0 = Reverse            ;  1 = Direct
  int am_mode;			// 0 = Manual             ;  1 = Automatic
  int algorithm;		// 0 = Ideal              ;  1 = Parallel
} pid[3];




/**************************************************************************** 
		Declare trend graph variables
****************************************************************************/

// This trend structure stores a fixed number of PV, SP, and OUT samples

#define TRENDWIDTH 120		// Set to 78 in order to fit an 80-colunm terminal
#define TRENDHEIGHT 25		// Set to 10 in order to fit a 24-row terminal

struct trendgraph
{
  float PV;
  float SP;
  float OUT;
} pen[TRENDWIDTH];

int trend_interval;		// The number of seconds between trend updates
float timebase;			// Measured in trend ticks per second





/**************************************************************************** 
		Declare DAQ variable arrays
****************************************************************************/

// The following fourteen-element array stores all the "raw" floating-point
// values read from the DAQ's analog input channels (AIN0 through AIN13)
float ain[14];


// The following two-element array stores all the "raw" floating-point
// values to be written to the the DAQ's analog output channels (DAC0 and DAC1)
float aout[2];



/**************************************************************************** 
		Declare other global variables
****************************************************************************/

int looprun = 1;		// This global variable allows the program loop to execute 
			 // when equal to 1.  Setting this variable to 0 causes the
			 // program to cleanly exit.


int select_mode = 0;		// This global variable controls which parameter is being adjusted
			 // 0 = operating mode -- adjusting OUT in manual or SP in automatic
			 // 1 = adjust K_P
			 // 2 = adjust K_I
			 // 3 = adjust integral deadband
			 // 4 = adjust K_D
			 // 5 = adjust action (direct/reverse)
			 // 6 = adjust algorithm (ideal/parallel)
			 // 7 = adjust trend interval

int time_current;		// This is the current UNIX system time cast into integer format
int time_lastscan;		// This is the previous UNIX system time (sampled during the last scan)
int scan_count = 0;		// This is a running tally of how many scans have been executed
float scans_per_second = 1;	// This is an estimate of how many program scans occur each second


/****************************************************************************
			Function prototypes 
****************************************************************************/

// The purpose of a function prototype in ANSI standard C programming is to
// tell the compiler what arguments will be passed to each function, and what
// sort of value will be returned by that function.  "void" means no argument
// or value is passed/returned.

// Prototype for reading a floating-point variable from any Modbus device
float read_32float (int address, int byteswap);

// Prototype for reading any generic integer variable via Modbus
int read_16int (int address);

// Prototype for writing a floating-point variable to any Modbus device
int write_32float (float real, int address, int byteswap);

// Prototype for the function initializing all PID control parameters
int splash_screen (void);

// Prototype for the function initializing all PID control parameters
int set_defaults (void);

// Prototype for the analog input channel reading function    
int pid_position (int);

// Prototype for plotting static displays to the screen
int display_plot (int);

// Prototype for reading keyboard input
int keyboard_scan (int);

// Prototype for the PID tuning adjustment environment
int tuning_entry (void);

// Prototype for shifting trend data arrays and plotting trend data points
int trend_shift_plot (int);

// Prototype for capturing trend graph data
int capture_trend (void);

// Prototype for initializing the Modbus network connection
int open_modbus_connection (void);

// Prototype for closing the Modbus network connection
int close_modbus_connection (void);

// Prototype for initializing the ncurses screen environment
int open_ncurses (void);

// Prototype for closing the ncurses screen environment
int close_ncurses (void);
